/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: test.c,v 1.23 2007/09/27 07:37:21 zlb Exp $ */

#include "phg.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <signal.h>

int
main(int argc, char *argv[])
{
    SIMPLEX *e;
    GRID *g, *g1;
    DOF *w, *v;
    const char *fn = "tetra.dat";
    FLOAT vol, weight;

    phgInit(&argc, &argv);

    g = phgNewGrid(-1);
    if (argc > 1)
	fn = argv[1];
    if (!phgImport(g, fn, FALSE))
	phgError(1, "can't read file \"%s\".\n", fn);

    vol = 0.0;
    ForAllElements(g, e)
	vol += phgGeomGetVolume(g, e);
    phgPrintf("Volume = %lg\n", (double)vol);

#if 0
    phgRefineAllElements(g, 1);
    ForAllElements(g, e) e->mark = -1;
    phgCoarsenMarkedElements(g);

    w = phgDofNew(g, DOF_P0, 1, "weight", DofNoAction);
    ForAllElements(g, e)
	*DofElementData(w, e->index) = pow((FLOAT)e->index, 2);
    phgBalanceGrid(g, 1.0, 0, w, 1.0);
    phgDofFree(&w);

    weight = 0.;
    ForAllElements(g, e)
	weight += pow((FLOAT)GlobalElement(g, e->index), 2);
    phgInfo(0, "weight: %lf, number elements: %d\n", (double)weight, g->nleaf); 

    if (TRUE) {	/* test FP exception */
	int i;
	weight = 1.0;
	for (i = 0; i < 10000; i++)
	    weight *= 0.1;		/* underflow */
	printf("weight = %lg\n", (double)weight);
	weight = 1.0 / weight;		/* divide by zero */
	printf("weight = %lg\n", (double)weight);
	weight = 1.0;
	weight = sqrt(-weight);		/* invalid operation */
	printf("weight = %lg\n", (double)weight);
    }
#endif

    v = phgDofNew(g, DOF_P0, 1, "v@g", DofNoAction);
    phgDofSetDataByValue(v, 0.0);
    g1 = phgNewGrid(-1);
    if (!phgImport(g1, "cube.dat", FALSE))
	phgError(1, "can't read file \"%s\".\n", "cube.dat");
    w = phgDofNew(g1, DOF_P0, 1, "w@g1", DofNoAction);
    phgDofSetDataByValue(w, 0.0);
phgDofDump(v);
phgDofDump(w);
    phgDofFree(&v);
    phgDofFree(&w);
    phgFreeGrid(&g1);

    phgFreeGrid(&g);
    phgFinalize();

    return 0;
}
